export namespace event {

  // =================================== IEvent ===================================
  /**Базовый класс параметров события*/
  export class EventArgs {

  }

  // =============================== IEventHandler ===============================
  /** Интерфейс функции обратного вызова */
  export interface IEventHandler<T extends EventArgs> {
    (sender: any, args: T): void;
  }

  // =================================== IEvent ===================================
  /**Интерфейс события */
  export interface IEvent<T extends EventArgs> {

    /**Регистрирует функцию, которая будет вызвана при наступлении события */
    on(callback: IEventHandler<T>): void;

    /**Регистрирует функцию, которая будет вызвана лишь однажды. Затем функция автоматически снимается с регистрации */
    once(callback: IEventHandler<T>): void;

    /**Снимает с регистрации указанную функцию. Если она не указана - снимает все функции у данного события */
    off(callback?: IEventHandler<T>): void;

    /**Вызывает все зарегистрированные функции */
    trigger(sender: any, args: T): void;
  }

  // =================================== Event ====================================
  /** Базовая реализация события */
  export class Event<T extends EventArgs> implements IEvent<T> {

    // Массив слушателей
    private _listeners: ({ callback: IEventHandler<T>, once: boolean })[] = [];

    /**Регистрирует функцию, которая будет вызвана при наступлении события */
    on(callback: IEventHandler<T>, once: boolean = false): void {
      let listener = this._listeners || [];
      for (let i = listener.length - 1; i >= 0; i--) {
        if (listener[i].callback === callback) {
          return;
        }
      }
      listener.push({ callback: callback, once: once });
      this._listeners = listener;
    }

    /**Регистрирует функцию, которая будет вызвана лишь однажды. Затем функция автоматически снимается с регистрации */
    once(callback: IEventHandler<T>): void {
      this.on(callback, true);
    }

    /**Снимает с регистрации указанную функцию. Если она не указана - снимает все функции у данного события */
    off(callback?: IEventHandler<T>): void {
      if (!callback) {
        // Обнулю массив обратных вызовов
        this._listeners = [];
      }
      else {
        let listener = this._listeners || [],
          i: number;
        for (let i = listener.length - 1; i >= 0; i--) {
          if (listener[i].callback === callback) {
            listener.splice(i, 1);
          }
        }
      }
    }

    /**Вызывает все зарегистрированные функции */
    trigger(sender: any, params: T): void {
      let onceListeners: number[] = [];
      let listeners = this._listeners || [];

      for (let i = 0; i < listeners.length; i++) {
        listeners[i].callback.apply(this, arguments);
        if (listeners[i].once) {
          onceListeners.push(i);
        }
      }
      // Сниму с регистрации все обработчики "на один раз"
      for (let i = onceListeners.length - 1; i >= 0; i--) {
        listeners.splice(onceListeners[i], 1);
      }
    }
  }

  // ==============================================================================
  // ==============================================================================

  // ================================= UIEvent ====================================
  /** Событие от пользовательского интерфейса */
  export class UIEventArgs extends EventArgs {
    constructor(eventName: string, eventData: any) {
      super();
      this._eventName = eventName;
      this._eventData = eventData;
    }

    private _eventName: string;
    get eventName(): string {
      return this._eventName;
    }

    private _eventData: any;
    get eventData(): any {
      return this._eventData;
    }
  }

  // ================================= UIEvent ====================================
  /** Событие от пользовательского интерфейса */
  export class UIEvent extends Event<UIEventArgs> { }

  // ============================== INotifyUIEvent ================================
  /**Интерфейс реализации уведомления от пользовательского интерфейса */
  export interface INotifyUIEvent {
    uiEvent: UIEvent;
  }

  // ==============================================================================
  // ==============================================================================

  // ========================== PropertyChangedEventArgs ==========================
  /**Интерфейс реализации уведомления об изменении свойств объекта */
  export class PropertyChangedEventArgs extends EventArgs {
    constructor(propertyName: string) {
      super();
      this._propertyName = propertyName;
    }

    _propertyName: string;
    get propertyName(): string {
      return this._propertyName;
    }
  }

  // ============================== PropertyChangedEvent ==========================
  /** Событие изменения свойства объекта */
  export class PropertyChangedEvent extends Event<PropertyChangedEventArgs> { }

  // ============================ INotifyPropertyChanged ==========================
  /**Интерфейс реализации уведомления об изменении свойств объекта */
  export interface INotifyPropertyChanged {
    propertyChanged: PropertyChangedEvent;
  }

  // ==============================================================================
  // ==============================================================================
  // ========================== Маршрутизируемые события ==========================
  // ==============================================================================
  // ==============================================================================

  // ========================== Стратегия маршрутизации ===========================
  /**Стратегия маршрутизации */
  export enum RoutingStrategy {
    /**Перенаправленное событие использует восходящую стратегию, 
     * где экземпляр события перемещается вверх по дереву от источника события к корневому элементу */
    Bubble = 1,
    /**Перенаправленное событие использует нисходящую стратегию, 
     * где экземпляр события перемещается вниз по дереву от корневого элемента к исходному */
    Tunnel = 2
  }

  // ========================= Перенаправленное событие ===========================
  /**Перенаправленное событие */
  export class RoutedEvent {
    /**Конструктор маршрутизируемого события: 
     * eventName: string - наименование события;
     * routingStrategy: RoutingStrategy - стратегия маршрутизации;
     * ownerType: any - тип объекта владельца события;
    */
    constructor(eventName: string, routingStrategy: RoutingStrategy, ownerType: any) {
      this._eventName = eventName;
      this._globalIndex = GlobalEventManager.getNextAvailableGlobalIndex(this);
      this._ownerType = ownerType;
      this._routingStranegy = routingStrategy;
    }

    private _ownerType: any;
    public get ownerType(): any {
      return this._ownerType;
    }

    private _eventName: string;
    /**Наименование события */
    get eventName(): string {
      return this._eventName;
    }

    private _routingStranegy: RoutingStrategy;
    /**Стратегия маршрутизации */
    get strategy(): RoutingStrategy {
      return this._routingStranegy;
    }

    private _globalIndex: number;
    get globalIndex(): number {
      return this._globalIndex;
    }
  }

  // ===================== Данные перенаправляемых событий ========================
  /**Данные перенаправляемых событий */
  export class RoutedEventArgs extends EventArgs {
    constructor(routedEvent: RoutedEvent) {
      super();
      if (!routedEvent) {
        throw new Error('RoutedEventArgs.constructor - параметр routedEvent - должен быть задан');
      }
      if (!RoutedEvent.prototype.isPrototypeOf(routedEvent)) {
        throw new Error('RoutedEventArgs.constructor - параметр routedEvent - должен наследоваться от RoutedEventArgs');
      }
      this._routedEvent = routedEvent;
      this._source = this._originalSource = null;
    }

    private _routedEvent: RoutedEvent;
    /**Событие */
    get routedEvent(): RoutedEvent {
      return this._routedEvent;
    }

    private _handled: boolean;
    /**Признак обработанности события */
    get handled() {
      return this._handled;
    }
    set handled(value: boolean) {
      this._handled = value;
    }

    private _originalSource: any;
    /**Начальный объект отправки события */
    get origianlSource(): any {
      return this._originalSource;
    }

    private _source: any;
    /**Источник события */
    get source(): any {
      return this._source;
    }
    set source(value: any) {
      if (this._source === null && this._originalSource == null) {
        this._source = this._originalSource = value;
      }
      else {
        this._source = value;
      }
    }
  }

  /**Параметры маршрутизируемого события имеющие расширенные данные */
  export class DataRoutedEventArgs extends RoutedEventArgs {
    constructor(event: RoutedEvent, eventData: any) {
      super(event);
      this._eventData = eventData;
    }

    private _eventData: any;
    public get eventData(): any {
      return this._eventData;
    }
    public set eventData(value: any) {
      this._eventData = value;
    }
  }

  // ==============================================================================
  // ==============================================================================
  /** Соглашение обработчиков маршрутизируемых событий */
  export interface RoutedEventHandler extends IEventHandler<RoutedEventArgs> { }

  // ========================= Менеджер перенаправляумых событий ===========================
  /** Менеджер перенаправляумых событий 
   * Содержит набот статических методов, занимающихся регистрацией событий в общей инфраструктуре
  */
  export class EventManager {

    static routedEvents: Object = {};

    /**Регистрирует маршрутизируемое событие в системе:
     * eventName - наименование регистрируемого события вида ИмяКласса.ИмяСобытия;
     * routingStrategy - стратегия маршрутизации 1 - восходящая, 2 - нисходящая;
     * ownerType - тип владельца события
    // */
    static registerRoutedEvent(eventName: string, routingStrategy: RoutingStrategy, ownerType: any): RoutedEvent {
      // Если еще не зарегистрирован
      if (!EventManager.routedEvents.hasOwnProperty(eventName)) {
        // Зарегистрирую
        (EventManager.routedEvents as any)[eventName] = new RoutedEvent(eventName, routingStrategy, ownerType);
      }
      return (EventManager.routedEvents as any)[eventName] as RoutedEvent;
    }

    /**Возвращает массив зарегистрированных перенаплавляемых событий */
    static getEvents(): RoutedEvent[] {
      const retVal = new Array<RoutedEvent>();
      for (let eventName in EventManager.routedEvents) {
        retVal.push((EventManager.routedEvents as any)[eventName]);
      }
      return retVal;
    }

    /**Возвращает именованный экземпляр перенаправляемого события */
    static getEvent(eventName: string): RoutedEvent {
      if (!EventManager.routedEvents.hasOwnProperty(eventName)) {
        throw new Error('Событие ' + eventName + ' не зарегистрировано');
      }
      return (EventManager.routedEvents as any)[eventName];
    }
  }

  // ==============================================================================
  // ==============================================================================
  /**Глобальный медеджер маршрутизируемых событий */
  export class GlobalEventManager {

    private static events: Array<Object> = new Array<Object>();

    public static getNextAvailableGlobalIndex(event: Object): number {
      const index = GlobalEventManager.events.push(event);
      return index;
    }
  }

  // ==============================================================================
  // ==============================================================================
  export class RoutedEventHandlerInfo {
    constructor(handler: RoutedEventHandler, handledEventsToo: boolean) {
      this._handler = handler;
      this._handledEventsToo = handledEventsToo;
    }

    private _handler: RoutedEventHandler;
    private _handledEventsToo: boolean;

    /**Обработчик */
    public get handler(): RoutedEventHandler {
      return this._handler;
    }

    /**Возвращает признак продолжения обработки даже обработанных событий или нет */
    public get invokeHandledEventsToo(): boolean {
      return this._handledEventsToo;
    }

    public invoke(target: any, args: RoutedEventArgs) {
      if (args.handled && !this.invokeHandledEventsToo) {
        // console.log('SIMA RoutedEventHandlerInfo invoke НЕ ЗАПУСТИЛ!!!');
        return;
      }
      // console.log('SIMA RoutedEventHandlerInfo invoke', target, args);
      this._handler.apply(target, arguments);
    }

    public equals(handler: RoutedEventHandler): boolean {
      return this._handler === handler;
    }
  }

  // ==============================================================================
  // ==============================================================================
  /**Хранилище обработчиков маршрутизируемых событий 
   * Используется классами, которые хранят обработчики событий
  */
  export class EventHandlerStore {
    constructor() {
      this._handlers = new Array<Array<RoutedEventHandlerInfo>>();
    }

    /**Внутренний массив массивов, в которых хранится обработчики
     * Представляет собой массив массивов. 
     * Индексом внешнего массива являются глобальные индексы зарегистрированных маршрутизируемых событий
     * (RoutedEvent.globalIndex)
     */
    private _handlers: Array<Array<RoutedEventHandlerInfo>>;

    /**Добавляет обработчик (callback) маршрутизируемого события в коллекцию */
    public addHandler(event: RoutedEvent, handler: RoutedEventHandler, handledEventsToo: boolean) {
      let h = this._handlers[event.globalIndex];
      if (!h) {
        this._handlers[event.globalIndex] = new Array<RoutedEventHandlerInfo>();
      }
      this._handlers[event.globalIndex].push(new RoutedEventHandlerInfo(handler, handledEventsToo));
    }

    public getHandlers(routedEvent: RoutedEvent): Array<RoutedEventHandlerInfo> {
      if (!routedEvent) {
        throw new Error('EventHandlerStore getHandlers не задано маршрутизируемое событие');
      }
      return this._handlers[routedEvent.globalIndex];
    }

    /**Удаляет обработчик привязанный к маршрутизируемому событию */
    public removeHandler(event: RoutedEvent, handler: RoutedEventHandler): void {
      let h = this._handlers[event.globalIndex];
      if (!h) {
        return;
      }
      for (let i = h.length - 1; i >= 0; i--) {
        if (h[i].equals(handler)) {
          h.splice(i, 1);
          return;
        }
      }
    }

    /**Удаляет все обработчики для заданного события */
    public removeEvent(event: RoutedEvent): void {
      this.removeEventByIndex(event.globalIndex);
    }

    /**Удаляет обработчики по глобальному индексу события */
    private removeEventByIndex(index: number): void {
      let h = this._handlers[index];
      if (!h) {
        if (Array.isArray(h)) {
          for (let i = h.length - 1; i >= 0; i--) {
            h.splice(i, 1);
          }
        }
        this._handlers.splice(index, 1);
      }
    }

    /**Удаляет все элементы коллекции обработчиков */
    public removeAll(): void {
      for (let key in this._handlers) {
        this.removeEventByIndex(Number.parseInt(key));
      }
    }
  }
}