import * as React from 'react';
import * as ReactDOM from 'react-dom';
import { osp } from './OverlapStackPanel/OverlapStackPanel';
import { CommonUtils } from './CommonUtils';
import { event } from './core/Event';
import ApiClient from './core/ApiClient';

// Чтобы не материлось Гы Гы Гы
declare function require(moduleName: string): any;

export namespace ui {

  /**Класс, применяющийся для приема высоты и ширины в вызовах AutoSizer'а
   * в виде <AutoSize>{(sz: ui.AutoSizerParams)=>{...}}
   * с последующей передачей полученных данные, к примеру в стиль ...Style={{height: sz.height, width: sz.width}}
   */
  export class AutoSizerParams {
    height: number;
    width: number;
  }

  /**Перечисление определяет куда выводится форма */
  export enum OutputType {
    /**На панель инструментов */
    ToolbarPanel = 1,
    /**На основную панель */
    OverlapPanel = 2,
    /**На форму */
    Form = 3,
    /**На панель заголовка */
    OverlapDashboard = 4
  }

  /**Режимы задания значений для табличных элементов и (или) списков*/
  export enum ValueSelectionModes {
    /**Выбирается значение из узла по пути memberPath */
    value = 1,
    /**В качестве значение используется индекс списка */
    index = 2,
    /**Выбирается весь объект */
    object = 3
  }

  /** Параметры создания вида*/
  export interface ViewerParams {
    /**Ключ (имя вида) */
    viewerKey?: any;
    /** Тип вывода */
    outputType?: OutputType;
    /** Ключ для элемента, в случае создания набора элементов */
    key?: any;
    /** Дополнительные параметры, которые передаются в элемент через props */
    props?: any;
  }

  /**Определение функции по созданию viewer'а */
  export interface CreateViewerFunction<T extends IController> {
    (
      /**Контроллер */
      controller: T,
      /**Параметры создания вьювера */
      params: ViewerParams
    ): IViewer<T>;
  }

  export interface CreateElementFunction<T extends IController, V extends IViewer<T>> {
    (
      viewer: V,
      createComponentRef: (component: any) => void,
      params: ViewerParams
    ): JSX.Element;
  }

  export interface IValueSelectionMode {
    /**Режим задания значения */
    mode: ValueSelectionModes;

    /**Путь к узлу содержащему значение. 
     * В начальном варианте просто имя свойства объекта, дальше планируется использовать JSONPath для более сложных случаев 
     * в этом случае значение должно начинаться с $
     */
    memberPath: string;

    /** Множитель
     * 1 - строго одно значение
     * 0..1 - ноль или одно значение
     * 1..* - одно и более значений
     * 0..* или * - сколько угодно значений, даже 0
    */
    multiplicity: string;
  }

  /**Объект, с помощью которого задается режим выбора значения объекта */
  export class ValueSelectionMode implements IValueSelectionMode {
    constructor() {
      this._mode = ValueSelectionModes.value;
      this.memberPath = 'id';
      this.multiplicity = '0..1';
    }

    /**Режим задания значения */
    private _mode: ValueSelectionModes;
    get mode(): ValueSelectionModes {
      return this._mode;
    }
    set mode(value: ValueSelectionModes) {
      this._mode = value;
    }

    /**Путь к узлу содержащему значение. 
     * В начальном варианте просто имя свойства объекта, дальше планируется использовать JSONPath для более сложных случаев 
     * в этом случае значение должно начинаться с $
     */
    private _memberPath: string;
    get memberPath(): string {
      return this._memberPath;
    }
    set memberPath(value: string) {
      this._memberPath = value;
    }

    /** Множитель
     * 1 - строго одно значение
     * 0..1 - ноль или одно значение
     * 1..* - одно и более значений
     * 0..* или * - сколько угодно значений, даже 0
    */
    private _multiplicity: string;
    get multiplicity(): string {
      return this._multiplicity;
    }
    set multiplicity(value: string) {
      if (value !== '1'
        && value !== '0..1'
        && value !== '1..*'
        && value !== '0..*'
        && value !== '*'
      ) {
        throw new Error('Недопустимый множитель для режима задания значения поля');
      }
      this._multiplicity = value;
    }

    get isSingleValue(): boolean {
      return this.multiplicity === '0..1' || this.multiplicity === '1';
    }

    get isValueRequired(): boolean {
      return this.multiplicity === '1' || this.multiplicity === '1..*';
    }

    static fromObject(obj: any): ValueSelectionMode {
      let v = new ValueSelectionMode();
      if (obj.hasOwnProperty('mode')) {
        if (obj.mode === 'value') {
          v.mode = ValueSelectionModes.value;
        }
        else if (obj.mode === 'object') {
          v.mode = ValueSelectionModes.object;
        }
        else if (obj.mode === 'index') {
          v.mode = ValueSelectionModes.index;
        }
        else {
          v.mode = obj.mode;
        }
      }
      if (obj.hasOwnProperty('memberPath')) {
        v.memberPath = obj.memberPath;
      }
      if (obj.hasOwnProperty('multiplicity')) {
        v.multiplicity = obj.multiplicity;
      }

      if (v.mode === ValueSelectionModes.value && (!v.memberPath || v.memberPath.toString().length === 0)) {
        throw new Error('Для режима выбора значения value необходимо задавать memberPath');
      }
      return v;
    }
  }

  // =============================== ITitled ======================================
  /**Интерфейс для объектов имеющих заголовок */
  export interface ITitled {
    /**Заголовок */
    title: string;
  }

  // =============================== IViewed ======================================
  /**Интерфейс отображаемых классов. Тех, кто может предоставить компонент отображения */
  export interface IViewed<T extends IController> {
    getViewer(params?: ViewerParams): IViewer<T>;
  }

  // =============================== IController ======================================
  /**Интерфейс контроллера */
  export interface IController {
    /**Отображаемое имя контроллера */
    controllerName: string;

    /**Контроллер владелец */
    ownerController: IController;

    /**Возвращает значение параметра переданного в конструктор контроллера через params */
    getCreationParamValue<PT>(paramName: string): PT;

    /**Добавление обработчика маршрутизируемого события */
    addHandler(routedEvent: event.RoutedEvent, handler: event.RoutedEventHandler, handledEventsToo: boolean): void;

    /**Удаление обработчика маршрутизируемого события */
    removeHandler(routedEvent: event.RoutedEvent, handler: event.RoutedEventHandler): void;

    /**Бросает событие */
    raiseEvent(routedEventArgs: event.RoutedEventArgs): void;

    /**Ищет по дереву класс содержащий контроллер перекрываемых панелей и возвращает его */
    getOverlapStackPanels(): osp.IOverlapStackPanels;
  }

  // =============================== IActionContainer =================================
  /**Интерфейс контейнера действий */
  export interface IActionContainer {
    /**Функция, возвращающая массив объектов действий класса */
    getActions(params?: Object): IAction[];
  }

  // =============================== IViewerProps =================================
  /**Интерфейс контекста данных */
  export interface IViewerProps<T extends IController> {
    controller: T;
    creationParams: any;
    viewer: IViewer<T>;
  }

  // // =========================== IModuleManager ===================================
  // export interface IModuleManager extends IController {

  // }

  // =============================== IModule ======================================
  /**Интерфейс модуля */
  export interface IModule extends IController, IViewed<IModule> {

    /**Id модуля */
    id: number;

    /**Имя модуля */
    name: string;

    /**Описание модуля */
    description: string;

    /**Класс ассоциированный с модулем
     * Если строка - ищется класс в каталоге classes как json или ts файл
     * Если объект - определение класса
     * Если функция - ожидается функция, которая вернет объект реализующий интерфейс IClass
     */
    class?: Object | ((params?: Object) => IClass) | string;

    /**Вид для класса. Актуально в том случае, когда class проинициализирован именем класса */
    keyView?: string;

    /**Возвращает определение класса (экземпляр класса) для данного модуля 
     * полученного по имени класса или из структурые его определения
    */
    getClass(params?: Object): IClass;
  }

  // =============================== IClass =======================================
  /**Интерфейс класса */
  export interface IClass extends IController, IActionContainer, ITitled, IViewed<IClass>, event.INotifyPropertyChanged, osp.IOverlapPanelController {
    /**Определение полей класса */
    fields: IField[] | ((params?: Object) => IField[]);

    /**Определение действий класса */
    actions: IAction[] | ((params?: Object) => IAction[]);

    /**Функция, возвращающая массив объектов полей класса */
    getFields(params?: Object): IField[];

    /**Функция, возвращающая массив объектов действий класса */
    getActions(params?: Object): IAction[];

    /**Ключ вида для определения класса */
    keyView: string;
  }

  // ================================== IField ====================================
  /**Интерфейс поля класса */
  export interface IField extends IController, ITitled, IViewed<IField>, event.INotifyPropertyChanged {

    /**Имя свойства класса*/
    beanPropertyName: string;

    /**Порядок следования поля */
    order: number;

    /**Признак видимости (по умолчанию видим) */
    // ПОКА НЕ ПОНЯТНО ЧТО С ЭТИМ ДЕЛАТЬ visible?: boolean;

    /**Указание на класс, который определяет содержимое настройки данного поля*/
    content: Object | ((params?: Object) => IClass) | string;

    /**Тип поля в "понимании" jsonSchema */
    // jsonType: any;

    /**Тип поля в "понимании" web service'а */
    // wsType: any;

    /**Тип данных */
    typeField: any;

    /**Признак того, является ли значение множественным  */
    isArray: boolean;

    /**Определение представления
     * Если строка - ищется tsx класс в каталоге views
     * Если функция - ожидается функция, которая вернет объект реализующий интерфейс IView и наследуемый от React.Component
     */
    viewer: ((params?: Object) => IViewer<IField>) | string;

    /**Признак отображения в панели инструментов */
    showInManagerPanel: boolean;

    /**Заголовок для вывода в панель управления */
    titleInManagerPanel: string;

    /**Разметка */
    propertiesLayout: string;

    /**Дополнительные свойства поля */
    propertiesField: any;

    /**Получение реального объекта содержимого */
    getContent(params?: Object): ContentClass;

    /**Определение панелей поля */
    panels: IPanel[] | ((params?: Object) => IPanel[]);

    /**Функция, возвращающая массив объектов панелей */
    getPanels(params?: Object): IPanel[];

    valueSelectionMode: ValueSelectionMode;
    /**Значение поля */
    value: any;
  }

  // ================================== IPanel ====================================
  /**Интерфейс действия */
  export interface IPanel extends IController, IActionContainer, IViewed<IPanel> {

    /**Наименование панели */
    name: string;

    /**Порядок следования действия */
    order: number;

    /**Определение представления
     * Если строка - ищется tsx класс в каталоге views
     * Если функция - ожидается функция, которая вернет объект реализующий интерфейс IView и наследуемый от React.Component
     */
    viewer: ((params?: Object) => IViewer<IField>) | string;

    /**Определение действий панели */
    actions: IAction[] | ((params?: Object) => IAction[]);

    /**Функция, возвращающая массив объектов действий */
    getActions(params?: Object): IAction[];

    /**Определение класса контроллера для данной панели */
    controller: ((params?: Object) => IClass) | string;

    /**Функция, возвращающая класс контроллера для заданной панели */
    getController(params?: Object): IClass;
  }

  // ================================== IAction ===================================
  /**Интерфейс действия */
  export interface IAction extends IController, ITitled, IViewed<IAction>, event.INotifyPropertyChanged {
    /**Признак срабатывания на двойной клик */
    doubleClick: boolean;

    /**Строка - путь к файлу картинки !!!!!! ОТСУТСТВУЕ В java*/
    icon: string;

    /**Событие - наименование бросаемого события !!!!!! ОТСУТСТВУЕ В java*/
    action: string;

    /**Порядок следования действия */
    order: number;

    /**Признак вывода действия на панель */
    showInManagerPanel: boolean;

    /**Признак вывода действия на панель окна, будут выводиться на Dashboard заголовка окна */
    showInPanel: boolean;

    /**Подсказка */
    hint: string;
  }

  // ================================== IViewer ===================================
  /**Интерфейс вида. По сути это фабрика элементов реакта */
  export interface IViewer<T extends IController> extends INotifyComponentChanged {
    /**Определение элемента отображения
     * либо строка, сожержащая имя tsx класса
     * либо функция фабрика, принимающая:
     *    owner - класс контроллера
     *    createComponentRef - функция обратного вызова, которую необходимо будет передать в качестве свойства props.ref при создании JSX элемента
     *    params - набор свойств, для создания
     */
    // element: ((owner: T, createComponentRef: (component: any) => void, params: ViewerParams) => JSX.Element) | string;
    // element: CreateViewerFunction<T> | string;
    element: CreateElementFunction<T, IViewer<T>> | string;

    controller: T;

    /**Параметры создания */
    viewerParams: ViewerParams;

    /**Функция принимающая набор параметров и инициирующая создание JSX элемента, определенного в element */
    // getElement(params?: Object): JSX.Element;
    getElement(): JSX.Element;
    /**Функция, возвращающая реальный React компонент, соответствующий element. */
    getComponent(): any;
  }

  // ============================== IApiDataSource ================================
  /**Интерфейс api источника данных */
  export interface IApiDataSource {

    /**Содержит имя сервиса данных */
    serviceName: string;

    /**Объявляет универсальный метод доступа к методам сервиса */
    call<T>(method: string, ...params: any[]): Promise<T>;
  }

  // ==============================================================================
  // =================================== События ==================================
  // ==============================================================================

  /**Класс реализующий аргументы события ComponentChanged */
  export class ComponentChangedEventArgs extends event.EventArgs {
    constructor(component: any) {
      super();
      this._component = component;
    }

    private _component: any;
    get component(): any {
      return this._component;
    }
  }

  export class ComponentChangedEvent extends event.Event<ComponentChangedEventArgs> { }

  export interface INotifyComponentChanged {
    componentChanged: ComponentChangedEvent;
  }

  // ==============================================================================
  // =================================== События ==================================
  // ==============================================================================

  // ==============================================================================
  // =============================== EventRouteItem =============================
  /**Элемент маршрута
   * Реализуется здесь, т.к. используется контроллером для маршрутизации событий 
   * */
  class EventRouteItem {
    /**target - целевой объект, обработчик которого должен быть вызван;
     * handlerInfo - информация об обработчике (методе целевого объекта)
     */
    constructor(target: IController, handlerInfo: event.RoutedEventHandlerInfo) {
      if (!target) {
        throw new Error('target должен быть задан');
      }
      if (!handlerInfo) {
        throw new Error('handlerInfo должен быть задан');
      }
      this._target = target;
      this._handlerInfo = handlerInfo;
    }

    private _target: IController;
    private _handlerInfo: event.RoutedEventHandlerInfo;

    public invoke(args: event.RoutedEventArgs) {
      this._handlerInfo.invoke(this._target, args);
    }
  }

  // =============================== EventRoute =============================
  /**Вспомагательный класс маршрута */
  class EventRoute {
    constructor(routedEvent: event.RoutedEvent) {
      this._routedEvent = routedEvent;
      this._routeItems = new Array<EventRouteItem>();
    }

    private _routedEvent: event.RoutedEvent;
    private _routeItems: Array<EventRouteItem>;

    /**Количество элементов в маршруте */
    public get count(): number {
      return this._routeItems.length;
    }

    /**Добавляет указанный обработчик для указанной цели маршрута.;
     * target - Определяет целевой объект, чей обработчик должен добавляться к маршруту;
     * handler - Определяет обработчик, который добавляется к маршруту;
     * handledEventsToo - Указывает, выявляет ли прослушиватель события, которые уже обрабатывались
     */
    public add(target: IController, handler: event.RoutedEventHandler, handledEventsToo: boolean) {
      if (!target) {
        throw new Error('Целевой объект должен быть задан');
      }
      if (!handler) {
        throw new Error('Обработчик должен быть задан');
      }
      this._routeItems.push(new EventRouteItem(target, new event.RoutedEventHandlerInfo(handler, handledEventsToo)));
    }

    /**Событие маршрута */
    public get routedEvent(): event.RoutedEvent {
      return this._routedEvent;
    }

    /**Выполняет обработчики по маршруту */
    public invokeHandlers(source: IController, args: event.RoutedEventArgs): void {
      if (!source) {
        throw new Error('source должен быть задан');
      }
      if (!args) {
        throw new Error('args должен быть задан');
      }
      if (!args.routedEvent) {
        throw new Error('args.routedEvent должен быть задан');
      }
      // console.log('SIMA EventRoute invokeHandlers', source, args);
      if (args.routedEvent.strategy === event.RoutingStrategy.Bubble) {
        // console.log('SIMA Всплывающее событие');
        for (let idx = 0; idx < this._routeItems.length; idx++) {
          this._routeItems[idx].invoke(args);
        }
      }
      else if (args.routedEvent.strategy === event.RoutingStrategy.Tunnel) {
        // console.log('SIMA Туннелированное событие');
        for (let idx = this._routeItems.length - 1; idx >= 0; idx--) {
          this._routeItems[idx].invoke(args);
        }
      }
    }
  }

  // ==============================================================================
  // ==============================================================================

  // =============================== UIControllerBase =============================
  /**Базовый класс для классов контроллеров раздела UI */
  abstract class UIControllerBase<T extends IController> implements IController, IViewed<T>, event.INotifyPropertyChanged {
    /**конструктор
     * ownerController - контроллер владелец
     * objectDef - объект определения
     * params - набор дополнительных параметров создания
     */
    constructor(ownerController: IController, objectDef?: Object, params?: Object) {

      console.warn('SIMA UIControllerBase.ctor ================================================', this);

      this._eventHandlerStore = new event.EventHandlerStore();
      this._ownerController = ownerController;
      this._creationParams = params || {};
      // Проинициализирую
      // this._viewer = null;
      this._viewers = new Array<IViewer<any>>();

      if (objectDef) {
        this.initFromDef(this, objectDef);
        this.validate();
      }

      // Подпишемся на уведомление об изменении компонента
      // Предполагается, что this._viewer проинициализируется (может проинициализироваться) в рамках this.initFromDef 
      // if (this._viewer !== null) {
      //   this._viewer.componentChanged.on(this.onComponentChanged.bind(this));
      // }
      for (let key in this._viewers) {
        // Такой подход, а не через forEach позволит обходить ключи и чисто массива, т.е. целый числа
        // и ключи объекта, т.е. если key будет числом - вернет значение по ключу массива, если key будет строкой - вернется свойство объекта
        this._viewers[key].componentChanged.on(this.onComponentChanged.bind(this));
      }
    }

    _eventHandlerStore: event.EventHandlerStore;
    propertyChanged: event.PropertyChangedEvent = new event.PropertyChangedEvent();

    static previewDoubleClick: event.RoutedEvent = new event.RoutedEvent('UIControllerBase.previewDoubleClick', event.RoutingStrategy.Tunnel, UIControllerBase);
    static doubleClick: event.RoutedEvent = new event.RoutedEvent('UIControllerBase.doubleClick', event.RoutingStrategy.Bubble, UIControllerBase);

    // Защищенное поле хранящее определение вида
    // private _viewer: IViewer<T>;
    // Внимание!!! Объявляем как массив, но использовать буду как объект, т.е. в качестве индекса будут не числа, а строки
    private _viewers: Array<IViewer<T>>;

    private _ownerController: IController;
    get ownerController(): IController {
      return this._ownerController;
    }
    set ownerController(value: IController) {
      this._ownerController = value;
    }

    /**Наименование контроллера */
    get controllerName(): string {
      throw new Error('Для контроллера не переопределен геттер controllerName');
    }

    /**Набор параметров, переданных в конструктор объекта в виде params */
    _creationParams: Object;

    /**Возвращает "типизированное" значение параметра создания объекта */
    getCreationParamValue<PT>(paramName: string): PT {
      if (!this._creationParams
        || !this._creationParams.hasOwnProperty(paramName)) {
        return undefined as PT;
      }
      return (this._creationParams as any)[paramName] as PT;
    }

    /**Добавление обработчика маршрутизируемого события */
    addHandler(routedEvent: event.RoutedEvent, handler: event.RoutedEventHandler, handledEventsToo: boolean = false): void {
      this._eventHandlerStore.addHandler(routedEvent, handler, handledEventsToo);
    }

    /**Удаление обработчика маршрутизируемого события */
    removeHandler(routedEvent: event.RoutedEvent, handler: event.RoutedEventHandler): void {
      this._eventHandlerStore.removeHandler(routedEvent, handler);
    }

    /**Бросает событие */
    raiseEvent(args: event.RoutedEventArgs): void {
      if (!args) {
        throw new Error('UIControllerBase.raiseEvent параметр args должен быть определен');
      }
      if (!event.RoutedEventArgs.prototype.isPrototypeOf(args)) {
        throw new Error('UIControllerBase.raiseEvent параметр args должен наследоваться от event.RoutedEventArgs');
      }
      if (!args.routedEvent) {
        throw new Error('UIControllerBase.raiseEvent в параметре args не указано маршрутизируемое событие');
      }
      UIControllerBase.raiseEventImpl(this, args);
    }

    /**Реализует процесс доставки события */
    private static raiseEventImpl(sender: IController, args: event.RoutedEventArgs) {
      args.source = sender;
      // Получу объект маршрута
      const route: EventRoute = UIControllerBase.getEventRote(sender, args);
      // Сформирую (заполню) маршрут обхода для события
      UIControllerBase.buildRoute(sender, route, args);
      route.invokeHandlers(sender, args);
    }

    /**Строит маршрут проходя от нисшего контроллера на самый верх*/
    private static buildRoute(sender: IController, route: EventRoute, args: event.RoutedEventArgs) {
      if (!args) {
        throw new Error('UIControllerBase.buildRoute параметр args должен быть определен');
      }
      if (!args.routedEvent) {
        throw new Error('UIControllerBase.buildRoute в параметре args не указано маршрутизируемое событие');
      }
      if (route.routedEvent !== args.routedEvent) {
        throw new Error('Событие маршрута не соответствует событию аргумента');
      }
      // Некоторая кривизна, но не хочется в интерфейс добавлять закрытое поле
      // поэтому через приведение и проверку свойства
      let ctrlr = sender;
      let store: event.EventHandlerStore;
      while (ctrlr) {
        store = ((ctrlr as any)._eventHandlerStore) as event.EventHandlerStore;
        if (store) {
          let list: Array<event.RoutedEventHandlerInfo> = store.getHandlers(args.routedEvent);
          if (list) {
            for (let idx = 0; idx < list.length; idx++) {
              let handlerInfo: event.RoutedEventHandlerInfo = list[idx];
              route.add(ctrlr, handlerInfo.handler, handlerInfo.invokeHandledEventsToo);
            }
          }
        }
        else {
          console.warn('SIMA UIControllerBase buildRoute не найдено поле _eventHandlerStore');
        }
        ctrlr = ctrlr.ownerController;
      }
    }

    /**Возвращает маршрут прохождения события */
    private static getEventRote(sender: IController, routedEventArgs: event.RoutedEventArgs): EventRoute {
      const route: EventRoute = new EventRoute(routedEventArgs.routedEvent);
      // let owner: IController = this;
      // while (owner !== null) {
      //   route.push(new EventRoute(owner, null));
      //   owner = owner.ownerController;
      // }
      return route;
    }

    /**Ищет по дереву класс содержащий контроллер перекрываемых панелей и возвращает его */
    getOverlapStackPanels(): osp.IOverlapStackPanels {
      let controller: any = this;
      while (controller) {
        if (controller['overlapStackPanels']) {
          return controller['overlapStackPanels'];
        }
        controller = (controller as IController).ownerController;
      }
      return null;
    }

    /**Инициализирует объект из его определения 
     * me - ссылка на сам настраиваемый объект (т.е. тут прилетит this)
     * d - определение структуры класса
    */
    protected initFromDef(me: UIControllerBase<T>, d: any): void {
      for (let propName in d) {
        if (!me.addSpecial(me, d, propName)) {
          (me as any)[propName] = d[propName];
        }
      }
    }

    /**Проверяет на "особое" свойство, то, которое требует особой обработки" 
     * me - ссылка на сам настраиваемый объект (т.е. тут прилетит this)
     * def - определение структуры класса
     * propName - имя свойства из определения
    */
    protected addSpecial(me: any, def: any, propName: string): boolean {
      if (propName === 'viewer' && def && def.viewer) {
        // let viewerKey: any = 'defaultViewer';
        // this._viewers[viewerKey] = this.initInternalViewer(me, def.viewer);
        return false; /// Определение вида оставим включенным в определение объекта
      }
      return false;
    }

    /**Метод, вызываемый в тот момент когда происходит изменение состояния компонента
     * Компонент может быть создан, тогда прилетит на него ссылка
     * Компонент может быть разрушен, тогда прилетит null
     */
    protected onComponentChanged(sender: any, args: ComponentChangedEventArgs) {
      if (args && args.component && args.component.uiEvent) {
        // Подпишемся на получение уведомлений от UI компонента
        // !!! Отписываться от событий не надо, т.к. это произойдет автоматически
        // компонент сам отпишет все обработчики в момент обработки componentWillUnmount
        let event: event.UIEvent = <event.UIEvent>args.component.uiEvent;
        event.on(this.onUIEvent.bind(this));
      }
    }

    /** Пустой обработчик для событий от UI*/
    protected onUIEvent(sender: any, eventData: event.UIEventArgs) {
      // console.log('SIMA onUIEvent', sender, eventData);
    }

    /**Проверяет на корректность данных объекта после инициализации.
     * В случае выявления дефектов (неполноты заполненности) бросать exceptnon
     */
    protected validate() { }

    protected initInternalViewer(
      controller: T,
      // viewerDef: ((owner: T, params: ViewerParams) => IViewer<T>) | string,
      viewerDef: CreateViewerFunction<T> | string,
      params: ViewerParams
    ): IViewer<T> {

      let tp = typeof viewerDef;

      if (tp === 'function') {
        // return (<(controller: T, params: ViewerParams) => IViewer<T>>viewerDef)(controller, params);
        return (<CreateViewerFunction<T>>viewerDef)(controller, params);
      }
      else if (tp === 'string') {
        return new Viewer<T>(controller, <string>viewerDef, params);
      }
      return null;
    }

    abstract getBaseViewerName(params?: ViewerParams): string;

    /** Формирует ключ для поиска вида в коллекции видов */
    getViewerKey(params?: ViewerParams): any {
      let viewerKey: any;
      if (params && params.outputType) {
        let outputType = params.outputType;
        viewerKey = 'outputType:' + outputType;
      }
      else {
        viewerKey = 'defaultViewer';
      }
      return viewerKey;
    }

    /**Метода интерфейса IViewed */
    getViewer(params?: ViewerParams): IViewer<T> {

      params = params || {};
      if (!params.viewerKey) {
        params.viewerKey = this.getViewerKey(params);
      }

      console.warn('SIMA getViewer -------------------------------------------------------', params.viewerKey, this._viewers, this);

      if (!this._viewers[params.viewerKey]) {
        console.log('SIMA getViewer ----------------------------- 1');
        this._viewers[params.viewerKey] = this.initInternalViewer(this as any, this.getBaseViewerName(params), params);
        // Подпишемся на уведомление об изменении компонента
        this._viewers[params.viewerKey].componentChanged.on(this.onComponentChanged.bind(this));
      }
      console.log('SIMA getViewer ----------------------------- 2');
      return this._viewers[params.viewerKey];
    }
  }

  // ================================== Module ====================================
  /**Базовый класс Модуля */
  export class Module extends UIControllerBase<Module> implements IModule {

    constructor(ownerController: IController, moduleDef?: Object) {
      super(ownerController, moduleDef);
      this.keyView = '';
    }

    get controllerName(): string {
      return 'ui/ui$ui.Module';
    }

    getBaseViewerName(params?: ViewerParams): string {
      return 'components/viewers/ModuleViewer';
    }

    /**Id модуля */
    id: number;

    /**Имя модуля */
    name: string;

    /**Описание модуля */
    description: string;

    /**Класс ассоциированный с модулем
     * Если строка - ищется класс в каталоге classes как json или ts файл
     * Если объект - определение класса
     * Если функция - ожидается функция, которая вернет объект реализующий интерфейс IClass
     */
    class: Object | ((params?: Object) => IClass) | string;

    /**Наименование вида */
    keyView: string;

    // Приватное поле хранящее определение класса
    private _class: IClass;

    protected addSpecial(me: any, def: any, propName: string): boolean {
      if (propName === 'class') {
        // this._class = this.initInternalClass(def.class);
        return false;
      }
      return super.addSpecial(me, def, propName);
    }

    private initInternalClass(classDef: Object | (() => IClass) | string): IClass {
      if (!classDef) {
        throw new Error('Ожидается определение или объявление класса в модуле');
      }
      let tp = typeof classDef;
      if (tp === 'string') {
        // Указано имя класса, создаю из файла
        return CommonUtils.createObjectFromFile(Class, this, <string>classDef);
      } else if (tp === 'function') {
        // Указана функция фабрика
        return (classDef as ((ownerController: IController) => IClass))(this);
      } else if (tp === 'object') {
        // Определение класса находится в самом объявлении модуля
        return new Class(this, classDef);
      }
      return null;
    }

    /**Возвращает определение класса (экземпляр класса) для данного модуля 
     * полученного по имени класса или из структурые его определения
    */
    getClass(params?: Object): IClass {
      if (!this._class && this.class) {
        this._class = this.initInternalClass(this.class);
      }
      return this._class;
    }

  }

  // ================================== Class =====================================
  /**Базовый класс Класса (Class) */
  export class Class extends UIControllerBase<Class> implements IClass {

    constructor(ownerController: IController, clsDef?: Object, params?: Object) {
      super(ownerController, clsDef, params);
      // Проинициализирую поля класса
      if (!this.title) this.title = '';
      if (!this._fields) this._fields = new Array<IField>();
      if (!this._actions) this._actions = new Array<IAction>();
    }

    get controllerName(): string {
      return 'ui/ui$ui.Class';
    }

    getBaseViewerName(params?: ViewerParams): string {
      return 'components/viewers/ClassViewer';
    }

    // propertyChanged: event.PropertyChangedEvent = new event.PropertyChangedEvent();

    /**Заголовок */
    title: string;

    /**Определение полей класса */
    fields: IField[] | ((params?: Object) => IField[]);

    /**Определение действий класса */
    actions: IAction[] | ((params?: Object) => IAction[]);

    /**Ключ вида для определения класса */
    keyView: string;

    /**Внутреннее поле для хранения коллекции объектов IField */
    private _fields: IField[];

    /**Внутреннее поле для хранения коллекции действий IField */
    private _actions: IAction[];

    /**Функция, возвращающая массив объектов полей класса */
    getFields(params?: Object): IField[] {

      if (!this._fields && this.fields && this.fields.length > 0) {
        this.addFields(this.fields);
      }

      let outputType: OutputType = OutputType.Form;
      if (params && params.hasOwnProperty('outputType')) {
        outputType = (params as any)['outputType'];
      }
      else {
        outputType = this.getCreationParamValue<OutputType>('outputType');
      }

      let fields = this._fields || [];
      if (outputType === OutputType.ToolbarPanel) {
        return fields.filter((field, index, arr) => { return field.showInManagerPanel === true; });
      }
      return fields;
    }

    getActions(params?: Object): IAction[] {

      if (!this._actions && this.actions && this.actions.length > 0) {
        this.addActions(this.actions);
      }

      let outputType: OutputType = OutputType.Form;
      if (params && params.hasOwnProperty('outputType')) {
        outputType = (params as any)['outputType'];
      }
      else {
        outputType = this.getCreationParamValue<OutputType>('outputType');
      }

      let actions = this._actions || [];
      if (outputType === OutputType.ToolbarPanel) {
        return actions.filter((action, index, arr) => { return action.showInManagerPanel === true; });
      }
      if (outputType === OutputType.OverlapDashboard) {
        return actions.filter((action, index, arr) => { return action.showInPanel === true; });
      }
      return actions;
    }

    protected addSpecial(me: any, def: any, propName: string) {
      if (propName === 'fields') {
        // this.addFields(def.fields);
        return false;
      }
      if (propName === 'actions') {
        // this.addActions(def.actions);
        return false;
      }
      return super.addSpecial(me, def, propName);
    }

    private addFields(fields: any): void {
      if (!Array.isArray(fields)) {
        throw new Error('Определение полей для класса должно являться массивом');
      }
      if (!this._fields) {
        this._fields = new Array<IField>(fields.length);
      }

      let params: any = new Object();
      let outputType: OutputType = this.getCreationParamValue<OutputType>('outputType');

      if (outputType !== undefined) {
        params.outputType = outputType;
      }
      for (let i = 0; i < fields.length; i++) {
        let fldDef = fields[i];
        let fld = new Field(this, fldDef, params);
        this._fields[i] = fld;
      }
    }

    private addActions(actions: any): void {
      if (!Array.isArray(actions)) {
        throw new Error('Определение действий для класса должно являться массивом');
      }
      if (!this._actions) {
        this._actions = new Array<IAction>(actions.length);
      }

      let params: any = new Object();
      let outputType: OutputType = this.getCreationParamValue<OutputType>('outputType');

      if (outputType !== undefined) {
        params.outputType = outputType;
      }
      for (let i = 0; i < actions.length; i++) {
        let acnDef = actions[i];
        // let acn = new Action(this, acnDef, params);
        let acn = CommonUtils.createAction(this, acnDef, params);
        if (acn) {
          this._actions[i] = acn;
        }
        else {
          console.warn('SIMA ui.Class.addActions. Для класса не создано действие по описанию', acnDef);
        }
      }
    }

    /* IOverlapPanelController */

    canClose(panelKey: string): boolean {
      return true;
    }

    protected _overlapStackPanels: osp.IOverlapStackPanels;

    get overlapStackPanels(): osp.IOverlapStackPanels {
      return this._overlapStackPanels;
    }
    set overlapStackPanels(value: osp.IOverlapStackPanels) {
      this._overlapStackPanels = value;
    }

    /* IOverlapPanelController */
    /**Метод интерфейса IOverlapPanelController, возвращающий панель управления контроллером */
    getDashboard(): JSX.Element {
      return this.getViewer({ outputType: OutputType.OverlapDashboard }).getElement();
    }
  }

  // =================================== Field ====================================
  /**Базовая реализация класса Поля (Field)*/
  export class Field extends UIControllerBase<Field> implements IField {

    constructor(ownerController: IClass, fldDef?: Object, params?: Object) {
      super(ownerController, fldDef, params);

      this.valueSelectionMode = null;

      if (!this._content) {
        this._content = null;
      }
    }

    get controllerName(): string {
      return 'ui/ui$ui.Field';
    }

    getBaseViewerName(params?: ViewerParams): string {
      let viewerKey = this.getViewerKey(params);
      if (viewerKey === 'defaultViewer' && this.viewer && typeof this.viewer === 'string') {
        return this.viewer as string;
      }
      return 'components/viewers/FieldViewer';
    }

    /**Класс содержимого */
    private _content: ContentClass;

    /**Заголовок поля */
    title: string;

    /**Имя свойства класса*/
    beanPropertyName: string;

    /**Порядок следования поля */
    order: number;

    /**Признак видимости (по умолчанию видим) */
    // ПОКА НЕ ПОНЯТНО ЧТО С ЭТИМ ДЕЛАТЬ 
    // visible?: boolean;

    /**Указание на класс, который определяет содержимое настройки данного поля*/
    content: Object | ((params?: Object) => IClass) | string;

    /**Тип поля в "понимании" jsonSchema */
    // jsonType: any;

    /**Тип поля в "понимании" web service'а */
    // wsType: any;

    /**Тип данных поля */
    typeField: any;

    /**Признак того, является ли значение множественным  */
    isArray: boolean;

    /**Определение представления
     * Если строка - ищется tsx класс в каталоге views
     * Если функция - ожидается функция, которая вернет объект реализующий интерфейс IView и наследуемый от React.Component
     */
    viewer: ((params?: Object) => IViewer<Field>) | string;

    /**Признак отображения в панели инструментов */
    showInManagerPanel: boolean;

    /**Заголовок для вывода в панель управления */
    titleInManagerPanel: string;

    /**Разметка */
    propertiesLayout: string;

    /**Дополнительные свойства поля */
    propertiesField: any;

    /** Обработчик событий поступающий от UI*/
    protected onUIEvent(sender: any, eventData: event.UIEventArgs) {
      console.log('SIMA Field onUIEvent', sender, eventData);
    }

    protected addSpecial(me: any, def: any, propName: string) {
      if (propName === 'content') {
        // this._content = this.initInternalContentClass(def.content); // Будет инициализироваться при запросе класса
        return false;
      }
      if (propName === 'panels') {
        // this.addPanels(def.panels);
        return false;
      }
      if (propName === 'valueSelectionMode') {
        this.valueSelectionMode = ValueSelectionMode.fromObject(def.valueSelectionMode);
        return true;
      }
      return super.addSpecial(me, def, propName);
    }

    private initInternalContentClass(classDef: Object | (() => IClass) | string): ContentClass {
      if (!classDef) {
        return;
      }
      let tp = typeof classDef;
      if (tp === 'string') {
        // Указано имя класса, создаю из файла
        return CommonUtils.createObjectFromFile(ContentClass, this, <string>classDef);
      } else if (tp === 'function') {
        // Указана функция фабрика
        return (classDef as (() => ContentClass))();
      } else if (tp === 'object') {
        // Определение класса находится в самом объявлении модуля
        return new ContentClass(this, classDef);
      }
      return null;
    }

    /**Получение реального объекта содержимого */
    getContent(params?: Object): ContentClass {
      console.log('SIMA Field.getContent !!!!!!!!!!!!!!!!!!!', this);
      if (!this._content && this.content) {
        this._content = this.initInternalContentClass(this.content);
      }
      return this._content;
    }

    private addPanels(panels: any): void {
      if (!Array.isArray(panels)) {
        throw new Error('Определение панелей для поля должно являться массивом');
      }
      if (!this._panels) {
        this._panels = new Array<IPanel>(panels.length);
      }
      for (let i = 0; i < panels.length; i++) {
        let pnlDef = panels[i];
        this._panels[i] = new Panel(this, pnlDef);
      }
    }

    /**Определение панелей поля */
    panels: IPanel[] | ((params?: Object) => IPanel[]);

    /**Внутреннее поле для хранения коллекции объектов IPanel */
    private _panels: IPanel[];

    /**Возвращает массив объектов панелей */
    getPanels(params?: Object): IPanel[] {
      if (!this._panels && this.panels && this.panels.length > 0) {
        this.addPanels(this.panels);
      }
      return this._panels || [];
    }

    private _valueSelectionMode: ValueSelectionMode;
    get valueSelectionMode(): ValueSelectionMode {
      return this._valueSelectionMode;
    }
    set valueSelectionMode(value: ValueSelectionMode) {
      if (this._valueSelectionMode !== value) {
        this._valueSelectionMode = value;
        this.propertyChanged.trigger(this, new event.PropertyChangedEventArgs('valueSelectionMode'));
      }
    }

    private _value: any;
    get value(): any {
      return this._value;
    }
    set value(value: any) {
      if (JSON.stringify(this._value) === JSON.stringify(value)) {
        return;
      }
      console.log('SIMA Field set Value', this._value, value);
      this._value = value;
      this.propertyChanged.trigger(this, new event.PropertyChangedEventArgs('value'));
    }
  }

  // =================================== Panel ====================================
  /**Базовый класс Панель (Panel)*/
  export class Panel extends UIControllerBase<Panel> implements IPanel {

    constructor(ownerController: IField, pnlDef?: Object, params?: Object) {
      super(ownerController, pnlDef, params);
    }

    get controllerName(): string {
      return 'ui/ui$ui.Panel';
    }

    getBaseViewerName(params?: ViewerParams): string {
      let viewerKey = this.getViewerKey(params);
      if (viewerKey === 'defaultViewer' && this.viewer && typeof this.viewer === 'string') {
        return this.viewer as string;
      }
      return 'components/viewers/PanelViewer';
    }

    protected addSpecial(me: any, def: any, propName: string) {
      if (propName === 'actions') {
        // this.addActions(def.actions);
        return false;
      }
      return super.addSpecial(me, def, propName);
    }

    /**Наименование панели */
    name: string;

    /**Порядок следования действия */
    order: number;

    /**Определение представления
     * Если строка - ищется tsx класс в каталоге views
     * Если функция - ожидается функция, которая вернет объект реализующий интерфейс IView и наследуемый от React.Component
     */
    viewer: ((params?: Object) => IViewer<Field>) | string;

    private addActions(actions: any): void {
      if (!Array.isArray(actions)) {
        throw new Error('Определение действий для панели должно являться массивом');
      }
      if (!this._actions) {
        this._actions = new Array<IAction>(actions.length);
      }
      // console.log('SIMA Panel.addActions ################################', actions);
      for (let i = 0; i < actions.length; i++) {
        let acnDef = actions[i];
        // console.log('SIMA Panel.addActions ################################', acnDef);
        // let acn = new Action(this, acnDef);
        let acn = CommonUtils.createAction(this, acnDef, null);
        if (acn) {
          this._actions[i] = acn;
        }
        else {
          console.warn('SIMA ui.Panel.addActions. Для панели не создано действие по описанию', acnDef);
        }
      }
    }

    /**Определение действий панели */
    actions: IAction[] | ((params?: Object) => IAction[]);

    /**Внутреннее поле для хранения коллекции объектов IAction */
    private _actions: IAction[];

    /**Функция, возвращающая массив объектов действий */
    getActions(params?: Object): IAction[] {
      if (!this._actions && this.actions && this.actions.length > 0) {
        this.addActions(this.actions);
      }
      return this._actions || [];
    }

    // addController(controller: any) {
    //   this.controller = controller;
    // }
    /**Определение класса контроллера для данной панели */
    controller: ((params?: Object) => IClass) | string;

    private _controller: IClass = null;

    /**Функция, возвращающая класс контроллера для заданной панели */
    getController(params?: Object): IClass {
      // console.log('SIMA Panel.getController', this);
      if (this.controller && this._controller === null) {
        if (typeof this.controller === 'string') {
          console.log('SIMA Panel.getController создаю для ', params, this.controller, this);
          this._controller = CommonUtils.createObjectFromFile(Class, this, <string>this.controller, params);
        }
        else {
          console.error('SIMA ui.Panel.getController не реализован способ создания контроллера панели не из строки');
        }
      }
      return this._controller;
    }
  }

  // =================================== Action ====================================

  export class Action extends UIControllerBase<IAction> implements IAction {

    constructor(ownerController: IController, acnDef?: Object, params?: Object) {
      super(ownerController, acnDef, params);
    }

    static previewDoAction: event.RoutedEvent = new event.RoutedEvent('Action.previewDoAction', event.RoutingStrategy.Tunnel, Action);
    static doAction: event.RoutedEvent = new event.RoutedEvent('Action.doAction', event.RoutingStrategy.Bubble, Action);

    get controllerName(): string {
      return 'ui/ui$ui.Action';
    }

    getBaseViewerName(params?: ViewerParams): string {
      let viewerKey = this.getViewerKey(params);
      if (viewerKey === 'defaultViewer' && this.viewer && typeof this.viewer === 'string') {
        return this.viewer as string;
      }
      return 'components/viewers/ActionViewer';
    }

    /**Признак срабатывания на двойной клик */
    doubleClick: boolean;

    /**Строка - путь к файлу картинки !!!!!! ОТСУТСТВУЕ В java*/
    icon: string;

    /**Событие - наименование бросаемого события */
    action: string;

    /**Порядок следования действия */
    order: number;

    /**Признак вывода действия на панель */
    showInManagerPanel: boolean;

    /**Признак вывода действия на панель окна, будут выводиться на Dashboard заголовка окна */
    showInPanel: boolean;

    /**Определение представления
     * Если строка - ищется tsx класс в каталоге views
     * Если функция - ожидается функция, которая вернет объект реализующий интерфейс IView и наследуемый от React.Component
     */
    viewer: ((params?: Object) => IViewer<IAction>) | string;

    /**Заголовок (hint) */
    title: string;

    /**Подсказка */
    hint: string;

    protected addSpecial(me: any, def: any, propName: string) {
      return super.addSpecial(me, def, propName);
    }

    getViewer(params?: ViewerParams): IViewer<IAction> {
      let v = super.getViewer(params);
      return v;
    }

    // addController(controller: any) {
    //   this.controller = controller;
    // }
    /**Определение класса контроллера для данного действия */
    // controller: ((params?: Object) => IClass) | string;

    // private _controller: IClass = null;

    /**Функция, возвращающая класс контроллера для заданного действия */
    // getController(params?: Object): IClass {
    //   if (this.controller && this._controller === null) {
    //     if (typeof this.controller === 'string') {
    //       this._controller = CommonUtils.createObjectFromFile(Class, this, <string>this.controller, params);
    //     }
    //     else {
    //       console.error('SIMA ui.Action.getController не реализован способ создания контроллера действия (Action) не из строки');
    //     }
    //   }
    //   return this._controller;
    // }
  }


  // =================================== Action ====================================
  /**Базовый класс Действия (Action)*/
  // export class Action extends UIControllerBase<Action> implements IAction {

  //   constructor(ownerController: IPanel | IClass, acnDef?: Object, params?: Object) {
  //     super(ownerController, acnDef, params);

  //     console.log('SIMA Action ++++++++++++++++++++', acnDef);
  //   }

  //   static previewDoAction: event.RoutedEvent = new event.RoutedEvent('Action.previewDoAction', event.RoutingStrategy.Tunnel, Action);
  //   static doAction: event.RoutedEvent = new event.RoutedEvent('Action.doAction', event.RoutingStrategy.Bubble, Action);

  //   get controllerName(): string {
  //     return 'ui/ui$Action';
  //   }

  //   getBaseViewerName(params?: ViewerParams): string {
  //     return 'components/viewers/ActionViewer';
  //   }

  //   /**Признак срабатывания на двойной клик */
  //   doubleClick: boolean;

  //   /**Строка - путь к файлу картинки !!!!!! ОТСУТСТВУЕ В java*/
  //   icon: string;

  //   /**Событие - наименование бросаемого события */
  //   action: string;

  //   /**Порядок следования действия */
  //   order: number;

  //   /**Признак вывода действия на панель */
  //   showInManagerPanel: boolean;

  //   /**Определение представления
  //    * Если строка - ищется tsx класс в каталоге views
  //    * Если функция - ожидается функция, которая вернет объект реализующий интерфейс IView и наследуемый от React.Component
  //    */
  //   viewer: ((params?: Object) => IViewer<Action>) | string;

  //   /**Заголовок (hint) */
  //   title: string;

  //   protected addSpecial(me: any, def: any, propName: string) {
  //     // if (propName === 'controller') {
  //     //   this.addController(def.controller);
  //     //   return false;
  //     // }
  //     return super.addSpecial(me, def, propName);
  //   }

  //   // addController(controller: any) {
  //   //   this.controller = controller;
  //   // }
  //   /**Определение класса контроллера для данного действия */
  //   controller: ((params?: Object) => IClass) | string;

  //   private _controller: IClass = null;

  //   /**Функция, возвращающая класс контроллера для заданного действия */
  //   getController(params?: Object): IClass {
  //     if (this.controller && this._controller === null) {
  //       if (typeof this.controller === 'string') {
  //         this._controller = CommonUtils.createObjectFromFile(Class, this, <string>this.controller, params);
  //       }
  //       else {
  //         console.error('SIMA ui.Action.getController не реализован способ создания контроллера действия (Action) не из строки');
  //       }
  //     }
  //     return this._controller;
  //   }
  // }

  export class MyAction extends Action {
    constructor(ownerController: IPanel | IClass, acnDef?: Object, params?: Object) {
      super(ownerController, acnDef, params);
    }

    static previewDoMyAction: event.RoutedEvent = new event.RoutedEvent('MyAction.previewDoMyAction', event.RoutingStrategy.Tunnel, MyAction);
    static doMyAction: event.RoutedEvent = new event.RoutedEvent('MyAction.doMyAction', event.RoutingStrategy.Bubble, MyAction);
  }

  // ==================================== Viewer ==================================
  /**Базовый класс Вида (Viewer)*/
  export class Viewer<T extends IController> implements IViewer<T> {

    constructor(
      /**Контроллер */
      controller: T,
      /**Определение viewer'а */
      // elementDef: CreateViewerFunction<T> | string,
      elementDef: CreateElementFunction<T, IViewer<T>> | string,
      /**Набор параметров */
      params: ViewerParams
    ) {
      this.ref_onComponentChanged = this.onComponentChanged.bind(this);
      this._controller = controller;
      // Сохраним определение компонента
      // Создавать будем при запросе
      this.element = elementDef;
      this.viewerParams = params;

      this._element = null;
    }

    // owner'ом является класс контроллера, такие как IClass, IField и т.д.
    protected _controller: T;

    viewerParams: ViewerParams;

    public get controller(): T {
      return this._controller;
    }

    // Определение JSX.Element'а
    // element: ((controller: T, createComponentRef: (component: any) => void, params?: Object) => JSX.Element) | string;
    // element: CreateViewerFunction<T> | string;
    element: CreateElementFunction<T, IViewer<T>> | string;

    // Созданный на основе определения JSX.Element
    protected _element: JSX.Element;
    // Компонент, соответствующий _element'у
    protected _component: any;

    /**Возвращает JSX.Element, который будет отображать содержимое контроллера */
    getElement(): JSX.Element {

      let params: any = {};

      if (this._element) {
        console.warn('SIMA getElement====================== 2', this._element, this);
        return null;
        // return this._element;
      }
      console.log('SIMA getElement====================== 1', this);

      let tp = typeof this.element;
      if (tp === 'string') {
      console.log('SIMA getElement====================== 3', this);
        this._element = CommonUtils.createElementFromFile<T>(this, this.ref_onComponentChanged, <string>this.element);
        if (!this._element) {
          console.error('SIMA Viewer<T>.getElement. Не удалось создать вьювер из файла', <string>this.element);
        }
      }
      else if (tp === 'function') {
        console.log('SIMA getElement====================== 4', this);
        // Определение построено на функции - вызовем ее
        // this._element = (<(owner: T, createComponentRef: (component: any) => void, params?: Object) => JSX.Element>this.element)(this._controller, this.ref_onComponentChanged, params);
        this._element = (<CreateElementFunction<T, IViewer<T>>>this.element)(this, this.ref_onComponentChanged.bind(this), params);
      }
      else {
        console.warn('Ожидается имя компонента');
      }
      return this._element;
    }

    getComponent() {
      return this._component;
    }

    /**Событие уведомляющее о создании / разрушении компонента */
    public componentChanged: ComponentChangedEvent = new ComponentChangedEvent();

    ref_onComponentChanged: (component: any) => void;

    /**Обработчик события от компонента 
     * Может прилетать ссылка на реальный component
     * Может прилетать null, в том случае когда компонент разрушен (разрушается)
    */
    onComponentChanged(component: any) {
      if (this._component !== component) {
        this._component = component;
        // Уведомлю о данном событии всех подписчиков, в нашем случае подписчиками будут контроллеры
        this.componentChanged.trigger(this, new ComponentChangedEventArgs(this._component));
      }
    }
  }

  /**SIMA Экспериментальный интерфейс призванный стандартизировать общение между источниками данных и видами */
  export interface IContent extends event.INotifyPropertyChanged {
    contentStructure: any;
    contentData: any;
  }

  // ================================= ContentClass ===============================
  /**Базовый класс Вида (Viewer)*/
  export class ContentClass extends Class implements IContent, IApiDataSource {

    constructor(ownerController: IController, clsDef?: Object, params?: Object) {
      super(ownerController, clsDef, params);
    }

    get controllerName(): string {
      return 'ui/ui$ui.ContentClass';
    }

    /**Наименование службы */
    serviceName: string;

    protected validate() {
      super.validate();
      if (!this.serviceName) {
        console.warn('SIMA для класса содержимого ' + this.title + ' не задано имя сервиса (свойство serviceName)');
      }
    }

    protected addSpecial(me: any, def: any, propName: string) {
      if (propName === 'serviceName') {
        this.serviceName = def[propName];
      }
      return super.addSpecial(me, def, propName);
    }

    private _filter: any;
    public get filter(): any {
      return this._filter;
    }
    public set filter(value: any) {
      if (value === this._filter) {
        return;
      }
      this._filter = value;
      this.propertyChanged.trigger(this, new event.PropertyChangedEventArgs('contentData'));
    }

    /**Базовая реализация обращения к методу веб сервиса */
    call<T>(method: string, ...params: any[]): Promise<T> {
      return ApiClient.call(this.serviceName + '/' + method, params);
    }

    /**Получение списка */
    getList<T>(...params: any[]): Promise<T> {
      return this.call<T>('getList', ...params);
      // .then((r: T) => {
      //   console.log('SIMA Content');
      //   throw new Error('Что-то пошло не так');
      // });
      // Через такое вот финт можно бросать ошибки и они будут по цепочке Promise отловлены
    }

    /**Удаление объекта из списка */
    delete(id: number) {
      return this.call<any>('delete', id);
    }

    /**Сохранение данных элемента списка */
    save(data: Object) {
      return this.call<any>('save', data);
    }

    /**Позволяет запросить у контроллера содержимого параметры для получения списка объектов */
    getListParams(): any {
      return {};
    }

    public get contentStructure(): any {
      throw new Error('Для класса содержимого ' + this.controllerName + ' свойство contentStructure не реализовано');
    }

    public get contentData(): any {
      throw new Error('Для класса содержимого ' + this.controllerName + ' свойство contentData не реализовано');
    }
  }

  // ==============================================================================
  // ==============================================================================
}